/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package org.sqlite;

import java.util.logging.Level;
import java.util.logging.Logger;
import org.sqlite.jdbc.TransactionType;
import org.sqlite.udf.Context;
import org.sqlite.schema.ColumnMetaData;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import org.sqlite.jdbc.JdbcConnection;
import org.sqlite.swig.SQLite3Constants;
import org.sqlite.types.SQLite3StmtPtrPtr;
import org.sqlite.udf.Function;
import org.sqlite.udf.ScalarFunction;
import static org.junit.Assert.*;

/**
 *
 * @author calico
 */
public class DatabaseTest {

    public DatabaseTest() {
    }

    @BeforeClass
    public static void setUpClass() throws Exception {
    }

    @AfterClass
    public static void tearDownClass() throws Exception {
    }

    @Before
    public void setUp() {
    }

    @After
    public void tearDown() {
    }

    @Test(expected = java.lang.UnsatisfiedLinkError.class)
    public void newThrowUnsatisfiedLinkError() throws SQLException {
        new Database(null, null).close();
    }

    @Test(expected = java.sql.SQLException.class)
    public void openFailed() throws SQLException, ClassNotFoundException {
        Class.forName("org.sqlite.Driver");
        final Database db = new Database("\\Program Files", null);
        db.close();
    }
    
    private static final String DATABASE = System.getProperty("user.dir") + "/test/unittest.db";
    
    private static Database newDatabase() throws ClassNotFoundException, SQLException {
        Class.forName("org.sqlite.Driver");
        return new Database(DATABASE, null);
    }
    
//    @Test
//    public void getURL() throws SQLException, ClassNotFoundException {
//        final Database db = newDatabase();
//        
//        assertEquals(SCHEME + DATABASE, db.getURL());
//        assertFalse((SCHEME.toUpperCase() + DATABASE).equals(db.getURL()));
//        assertFalse((SCHEME + DATABASE.toUpperCase()).equals(db.getURL()));
//
//        db.close();
//    }
    
    @Test
    public void getProductName() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        assertEquals("SQLite", db.getProductName());
        
        db.close();
    }
    
    @Test
    public void isClosed() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        assertFalse(db.isClosed());
        
        db.close();
        assertTrue(db.isClosed());
    }
    
    @Test
    public void transaction() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        assertTrue(db.getAutoCommit());
        
        db.beginTransaction(null);
        assertFalse(db.getAutoCommit());
        
        db.rollbackTransaction();
        assertTrue(db.getAutoCommit());
        
        db.beginTransaction(null);
        assertFalse(db.getAutoCommit());
        
        db.commitTransaction();
        assertTrue(db.getAutoCommit());
        
        db.close();
    }
    
    @Test
    public void busyTimeout() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        assertEquals(0, db.getBusyTimeout());
        
        db.setBusyTimeout(500);
        assertEquals(500, db.getBusyTimeout());
        
        db.setBusyTimeout(-500);
        assertEquals(0, db.getBusyTimeout());
        
        db.close();
    }
    
    @Test
    public void execute() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        String sql= null;
        db.execute(sql);
        
        sql= "";
        db.execute(sql);
        
        sql= "SELECT * FROM sqlite_master";
        db.execute(sql);
        
        sql= "SELECT * FROM sqlite_master WHERE type = ?";
        db.execute(sql);
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void executeSyntaxError() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        String sql= "SELECT * FROM";
        db.execute(sql);
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void executeTableNotFound() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        String sql= "SELECT * FROM _NOT_FOUND_TABLE_";
        db.execute(sql);
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void executeColumnNotFound() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        String sql = "SELECT * FROM sqlite_master WHERE _NOT_FOUND_COLUMN_ = ?";
        db.execute(sql);
        
        db.close();
    }
    
    @Test(expected = java.lang.NullPointerException.class)
    public void pragmaThrowNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        db.pragma(null);
        db.close();
    }
    
    @Test
    public void pragma() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        Statement stmt = null;
        
        // confirm default valie
        stmt = db.prepare("PRAGMA read_uncommitted");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(0, stmt.getInt(1));
        stmt.close();
        
        stmt = db.prepare("PRAGMA full_column_names");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(0, stmt.getInt(1));
        stmt.close();

        stmt = db.prepare("PRAGMA cache_size");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(2000, stmt.getInt(1));
        stmt.close();

        // execute PRAGMA
        final String[] commands
                = new String[] {
                        "read_uncommitted = on",
                        "full_column_names = on",
                        "cache_size = 9876",
                    };
        db.pragma(commands);
        
        // validate pragma() funcation.
        stmt = db.prepare("PRAGMA read_uncommitted");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(1, stmt.getInt(1));
        stmt.close();
        
        stmt = db.prepare("PRAGMA full_column_names");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(1, stmt.getInt(1));
        stmt.close();

        stmt = db.prepare("PRAGMA cache_size");
        assertEquals(SQLite3Constants.SQLITE_ROW, stmt.step());
        assertEquals(9876, stmt.getInt(1));
        stmt.close();
        
        db.close();
    }
    
    @Test(expected = java.lang.NullPointerException.class)
    public void prepareThrowSqlNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        final SQLite3StmtPtrPtr ppStmt = new SQLite3StmtPtrPtr();
        Statement stmt = db.prepare(null, ppStmt);
        stmt.close();
        ppStmt.delete();
        
        db.close();
    }
    
    @Test(expected = java.lang.NullPointerException.class)
    public void prepareThrowSQLite3StmtPtrPtrNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        Statement stmt = db.prepare("SELECT * FROM sqlite_master", null);
        stmt.close();

        db.close();
    }
    
    @Test
    public void prepareWithSQLite3StmtPtrPtr() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        final SQLite3StmtPtrPtr ppStmt = new SQLite3StmtPtrPtr();
        Statement stmt = db.prepare("SELECT * FROM sqlite_master", ppStmt);
        stmt.step();
        stmt.close();
        ppStmt.delete();
        
        db.close();
    }
    
    @Test(expected = java.lang.NullPointerException.class)
    public void prepareThrowNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        Statement stmt = db.prepare(null);
        stmt.close();

        db.close();
    }
    
    @Test
    public void prepare() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        
        Statement stmt = db.prepare("SELECT * FROM sqlite_master");
        stmt.step();
        stmt.close();
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void prepareMultiple() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        
        try {
            String sql = null;

            sql = "SELECT * FROM sqlite_master;"
                + "SELECT * FROM sqlite_temp_master;"
                + ";";
            List<Statement> stmts = db.prepareMultiple(sql);
            assertEquals(3, stmts.size());
            for (final Statement stmt : stmts) {
                stmt.step();
                stmt.close();
            }
        } finally {
            db.close();
        }
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void interruptThrowSQLException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        Statement stmt = db.prepare("SELECT * FROM sqlite_master");
        db.interrupt();
        stmt.step();
        stmt.close();
        
        db.close();
    }
    
    @Test
    public void interrup() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        Statement stmt = db.prepare("SELECT * FROM sqlite_master");
        db.interrupt();
        stmt.close();
        
        db.close();
    }
    
    @Test
    public void changes() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        db.execute(
                "CREATE TEMPORARY TABLE IF NOT EXISTS temp_tbl_1("
                    + "  ID INTEGER PRIMARY KEY"
                    + ", VALUE TEXT"
                + ")"
            );
        assertEquals(0, db.changes());
        
        db.beginTransaction(TransactionType.IMMEDIATE);
        Statement stmt = db.prepare("INSERT INTO temp_tbl_1(ID, VALUE) VALUES(@id, @value)");
        for (int i = 0; i < 5; ++i) {
            stmt.reset();
            stmt.bindInt(stmt.getParameterIndex("@id"), i);
            
            byte[] text = new byte[5];
            for (int j = 0; j < text.length; ++j) {
                text[j] = (byte) ('A' + i + j * 5);
            }
            stmt.bindText(stmt.getParameterIndex("@value"), new String(text));
            stmt.execute();
            assertEquals(1, db.changes());
        }
        stmt.close();
        db.commitTransaction();

        stmt = db.prepare("UPDATE temp_tbl_1 SET VALUE = lower(VALUE) WHERE ID BETWEEN @from AND @to");
        stmt.bindInt(stmt.getParameterIndex("@from"), 1);
        stmt.bindInt(stmt.getParameterIndex("@to"), 4);
        stmt.execute();
        assertEquals(4, db.changes());
        stmt.close();
        
        db.close();
    }
    
    @Test
    public void totalChanges() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        db.execute(
                "CREATE TEMPORARY TABLE IF NOT EXISTS temp_tbl_2("
                    + "  ID INTEGER PRIMARY KEY"
                    + ", VALUE TEXT"
                + ")"
            );
        assertEquals(0, db.totalChanges());

        db.execute(
                "CREATE TEMPORARY TABLE IF NOT EXISTS temp_tbl_2_bak("
                    + "  ID INTEGER PRIMARY KEY"
                    + ", VALUE TEXT"
                + ")"
            );
        assertEquals(0, db.totalChanges());

        db.execute(
                "CREATE TEMPORARY TRIGGER backup_temp_tbl_2 UPDATE OF VALUE ON temp_tbl_2"
                    + " FOR EACH ROW WHEN NOT EXISTS(SELECT * FROM temp_tbl_2_bak WHERE ID = OLD.ID)"
                    + " BEGIN"
                        + " INSERT INTO temp_tbl_2_bak VALUES(OLD.ID, OLD.VALUE);"
                    + " END;"
            );
        assertEquals(0, db.totalChanges());
        
        db.beginTransaction(TransactionType.IMMEDIATE);
        Statement stmt = db.prepare("INSERT INTO temp_tbl_2(ID, VALUE) VALUES(@id, @value)");
        int totalChanges = 0;
        for (int i = 0; i < 5; ++i) {
            stmt.reset();
            stmt.bindInt(stmt.getParameterIndex("@id"), i);
            
            byte[] text = new byte[5];
            for (int j = 0; j < text.length; ++j) {
                text[j] = (byte) ('A' + i + j * 5);
            }
            stmt.bindText(stmt.getParameterIndex("@value"), new String(text));
            stmt.execute();
            assertEquals(1, db.changes());
            totalChanges += db.changes();
            assertEquals(totalChanges, db.totalChanges());
        }
        stmt.close();
        db.commitTransaction();

        stmt = db.prepare("UPDATE temp_tbl_2 SET VALUE = lower(VALUE) WHERE ID BETWEEN @from AND @to");
        stmt.bindInt(stmt.getParameterIndex("@from"), 1);
        stmt.bindInt(stmt.getParameterIndex("@to"), 4);
        stmt.execute();
        assertEquals(4, db.changes());
        totalChanges += (db.changes() * 2); // Triger
        assertEquals(totalChanges, db.totalChanges());
        stmt.close();
        
        db.close();
    }
    
    @Test
    public void lastInsertRowId() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        db.execute(
                "CREATE TEMPORARY TABLE IF NOT EXISTS temp_tbl_3("
                    + "  ID INTEGER PRIMARY KEY"
                    + ", VALUE TEXT"
                + ")"
            );
        
        db.beginTransaction(TransactionType.IMMEDIATE);
        Statement stmt = db.prepare("INSERT INTO temp_tbl_3(VALUE) VALUES(?1)");
        for (int i = 0; i < 5; ++i) {
            stmt.reset();
            stmt.bindText(1, "qwerty");
            stmt.execute();
            assertEquals((long) (i + 1), db.lastInsertRowId());
        }
        stmt.close();
        db.commitTransaction();
        
        db.close();
    }
    
    @Test(expected = java.lang.NullPointerException.class)
    public void addStatementThrowNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        db.addStatement(null);
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void addStatementThrowSQLException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        final Statement stmt = new Statement(db, new SQLite3StmtPtrPtr());
        db.addStatement(stmt);
        
        db.close();
    }
    
    @Test
    public void addStatement() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        final Database db2 = newDatabase();

        final Statement stmt = new Statement(db, new SQLite3StmtPtrPtr());
        db2.addStatement(stmt);
        db2.removeStatement(stmt);
        
        db2.close();
        db.close();

        assertTrue(stmt.isClosed());        
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void removeStatementThrowNullPointerException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        final Statement stmt = new Statement(db, new SQLite3StmtPtrPtr());
        db.removeStatement(null);
        
        db.close();
    }
    
    @Test(expected = java.sql.SQLException.class)
    public void removeStatementThrowSQLException() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        final Database db2 = newDatabase();
        final Statement stmt = new Statement(db, new SQLite3StmtPtrPtr());
        db2.addStatement(stmt);

        final Statement stmt2 = new Statement(db, new SQLite3StmtPtrPtr());
        db2.removeStatement(stmt2);
        
        db2.close();
        db.close();
    }
    
    @Test
    public void removeStatement() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();

        final Statement stmt = new Statement(db, new SQLite3StmtPtrPtr());
        db.removeStatement(stmt);
        
        db.close();
        
        assertTrue(stmt.isClosed());
    }

    @Test
    public void getColumnMetaData() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        try {
            final Connection conn = new JdbcConnection(db, null);
            
            final java.sql.Statement stmt = conn.createStatement();
            String sql
                    = "CREATE TEMPORARY TABLE IF NOT EXISTS temp_tbl_1("
                        + "  ROWID INTEGER PRIMARY KEY AUTOINCREMENT"
                        + ", ID INTEGER NOT NULL"
                        + ", tVALUE TEXT DEFAULT '_blank' COLLATE NOCASE"
                        + ", rVALUE REAL"
                        + ", bVALUE BLOB"
                        + ", nVALUE NULL"
                        + ", tVALUE2 text DEFAULT ''"
                        + ", tVALUE3 VarChar DEFAULT NULL"
                    + ")";
            stmt.executeUpdate(sql);
        
            ColumnMetaData meta = db.getColumnMetaData(null, "temp_tbl_1", "ROWID");
            assertEquals("INTEGER", meta.dataType);
            assertEquals("BINARY", meta.collationSequenceName);
            assertFalse(meta.isNotNull);
            assertTrue(meta.isPrimaryKey);
            assertTrue(meta.isAutoIncrement);
            
            meta = db.getColumnMetaData(null, "temp_tbl_1", "ID");
            assertEquals("INTEGER", meta.dataType);
            assertEquals("BINARY", meta.collationSequenceName);
            assertTrue(meta.isNotNull);
            assertFalse(meta.isPrimaryKey);
            assertFalse(meta.isAutoIncrement);
            
            meta = db.getColumnMetaData(null, "temp_tbl_1", "tVALUE");
            assertEquals("TEXT", meta.dataType);
            assertEquals("NOCASE", meta.collationSequenceName);
            assertFalse(meta.isNotNull);
            assertFalse(meta.isPrimaryKey);
            assertFalse(meta.isAutoIncrement);
            
            
            conn.close();
            
        } finally {
            db.close();
        }
    }

    private static Connection newConnection(Database db) throws ClassNotFoundException, SQLException {
        return new JdbcConnection(db, null);
    }

    @Test
    public void createFunctions() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        try {
            final Function myFunc
                    = new ScalarFunction("myFunc") {
                            @Override
                            public void xFunc(Context ctx) throws SQLException {
                                System.out.println("called xFunc()!");
                                String val = ctx.getString(1);
                                ctx.result(val);
                            }
                        };
            assertFalse(myFunc.isRegistered());
            db.createFunction(myFunc);
            assertTrue(myFunc.isRegistered());
            
            final Connection conn = newConnection(db);
            final java.sql.Statement stmt = conn.createStatement();
            ResultSet rs = stmt.executeQuery("SELECT myFunc('TEST') LIMIT 1");
            assertTrue(rs.next());
            assertEquals("TEST", rs.getString(1));
            rs.close();
            
            stmt.close();
            conn.close();
            
        } finally {
            db.close();
        }
    }
    
    @Test
    public void close() throws ClassNotFoundException, SQLException {
        final Database db = newDatabase();
        try {
            // nothing
        } finally {
            new Thread() {
                @Override
                public void run() {
                    try {
                        db.close();
                        System.out.println("close database.");
                    } catch (SQLException ex) {
                        Logger.getLogger(DatabaseTest.class.getName()).log(Level.SEVERE, null, ex);
                    }
                }
            }.start();
        }
    }
}